import unittest
import queue

from vigi.utils.pub_sub import PubSub

class TestPubSub(unittest.TestCase):
    def test_single_subscription(self):
        """Test that a single subscriber receives messages."""

        pubsub = PubSub()
        queue = pubsub.subscribe()

        test_message = "Hello, World!"
        pubsub.publish(test_message)

        self.assertEqual(queue.get(), test_message)
        pubsub.unsubscribe(queue)


    def test_multiple_subscriptions(self):
        """Test that multiple subscribers receive messages."""

        pubsub = PubSub()
        queues = [pubsub.subscribe() for _ in range(3)]
        test_message = "Hello, Multiworld!"
        pubsub.publish(test_message)

        for queue in queues:
            self.assertEqual(queue.get(), test_message)
            pubsub.unsubscribe(queue)

    def test_unsubscribe(self):
        """Test that unsubscribing prevents receiving messages."""
        pubsub = PubSub()
        queue1 = pubsub.subscribe()
        queue2 = pubsub.subscribe()
        pubsub.unsubscribe(queue1)
        
        test_message = "Test Unsubscribe"
        pubsub.publish(test_message)

        # queue1 should not receive any messages since it's unsubscribed
        with self.assertRaises(queue.Empty):
            queue1.get(timeout=1)

        # queue2 should still receive the message
        self.assertEqual(queue2.get(), test_message)
        pubsub.unsubscribe(queue2)



import cv2
import numpy as np

camera = cv2.VideoCapture(0)

back_sub = cv2.createBackgroundSubtractorMOG2(varThreshold=100, detectShadows=True)

while True:
    success, original_frame = camera.read()  # Read a frame from the camera
    if not success:
        break

    # reduce the frame size to 50%
    original_frame = cv2.resize(original_frame, (0, 0), fx=0.5, fy=0.5)
    original_frame_img = original_frame.copy()
    # write the text "Original" on the frame
    cv2.putText(original_frame_img, "1. Original", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(original_frame_img, (0, 0), (original_frame_img.shape[1], original_frame_img.shape[0]), (100, 100, 100), 2)

    gray_frame = cv2.cvtColor(original_frame, cv2.COLOR_BGR2GRAY)
    gray_frame_img = cv2.cvtColor(gray_frame, cv2.COLOR_GRAY2BGR)
    # write the text "Gray" on the frame
    cv2.putText(gray_frame_img, "2. Gray", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(gray_frame_img, (0, 0), (gray_frame_img.shape[1], gray_frame_img.shape[0]), (100, 100, 100), 2)

    fg_mask = back_sub.apply(gray_frame)
    fg_mask_img = cv2.cvtColor(fg_mask, cv2.COLOR_GRAY2BGR)
    # write the text "Foreground Mask" on the frame
    cv2.putText(fg_mask_img, "3. Foreground Mask", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(fg_mask_img, (0, 0), (fg_mask_img.shape[1], fg_mask_img.shape[0]), (100, 100, 100), 2)

    min_thresh = 100
    _result, motion_mask = cv2.threshold(fg_mask, thresh = min_thresh, maxval = 255, type = cv2.THRESH_BINARY)
    motion_mask_img = cv2.cvtColor(motion_mask, cv2.COLOR_GRAY2BGR)
    # write the text "Thresholded Motion Mask" on the frame
    cv2.putText(motion_mask_img, "4. Thresholded Motion Mask", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(motion_mask_img, (0, 0), (motion_mask_img.shape[1], motion_mask_img.shape[0]), (100, 100, 100), 2)

    motion_mask_blured = cv2.medianBlur(motion_mask, ksize = 3)
    motion_mask_blured_img = cv2.cvtColor(motion_mask_blured, cv2.COLOR_GRAY2BGR)
    # write the text "Motion Mask Blured" on the frame
    cv2.putText(motion_mask_blured_img, "5. Motion Mask Blured", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(motion_mask_blured_img, (0, 0), (motion_mask_blured_img.shape[1], motion_mask_blured_img.shape[0]), (100, 100, 100), 2)

    kernel = np.array((21,21), dtype=np.uint8)
    motion_mask_morph_open = cv2.morphologyEx(motion_mask_blured, op = cv2.MORPH_OPEN, kernel = kernel, iterations = 3)

    motion_mask_morph_close = cv2.morphologyEx(motion_mask_morph_open, op = cv2.MORPH_CLOSE, kernel = kernel, iterations = 3)
    motion_mask_morph_open_close_img = cv2.cvtColor(motion_mask_morph_close, cv2.COLOR_GRAY2BGR)
    # write the text "Motion Mask Morph Close" on the frame
    cv2.putText(motion_mask_morph_open_close_img, "6. MORPH_OPEN and MORPH_CLOSE", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(motion_mask_morph_open_close_img, (0, 0), (motion_mask_morph_open_close_img.shape[1], motion_mask_morph_open_close_img.shape[0]), (100, 100, 100), 2)
    
    contours, _ = cv2.findContours(motion_mask_morph_close, mode = cv2.RETR_EXTERNAL, method = cv2.CHAIN_APPROX_SIMPLE)
    # draw the contours on the black image
    contours_img = np.zeros_like(motion_mask_morph_open_close_img)
    cv2.drawContours(contours_img, contours, -1, (255, 255, 255), 2)
    # contours_img = cv2.cvtColor(contours_img, cv2.COLOR_GRAY2BGR)
    # write the text "Contours" on the frame
    cv2.putText(contours_img, "7. Contours", (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.rectangle(contours_img, (0, 0), (contours_img.shape[1], contours_img.shape[0]), (100, 100, 100), 2)

    # filter by contour area
    cnt_area_thresh = 5000
    selected_countours = []
    detections = []
    for cnt in contours:
        
        area = cv2.contourArea(cnt)
        if area > cnt_area_thresh:
            x, y, w, h = cv2.boundingRect(cnt)
            selected_countours.append(cnt)
            detections.append([x, y, w, h])

    # draw the contours on the black image
    cv2.drawContours(contours_img, selected_countours, -1, (0, 0, 255), 2)
    # contours_img = cv2.cvtColor(contours_img, cv2.COLOR_GRAY2BGR)
    # write the text "Thresholded Contours" on the frame

    result_img = original_frame.copy()

    # bounding box
    for detection in detections:
        x, y, w, h = detection
        cv2.rectangle(result_img, (x, y), (x + w, y + h), (0, 255, 0), 2)
    
    cv2.putText(result_img, '8. Detections', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)

    # drow frame around the image
    cv2.rectangle(result_img, (0, 0), (result_img.shape[1], result_img.shape[0]), (100, 100, 100), 2)

    numpy_vertical = np.vstack((original_frame_img,
                                gray_frame_img,
                                fg_mask_img,
                                motion_mask_img))
    
    numpy_vertical2 = np.vstack((motion_mask_blured_img,
                                motion_mask_morph_open_close_img,
                                contours_img,
                                result_img))

    numpy_horizontal = np.hstack((numpy_vertical, numpy_vertical2))
    

    # show both the original and the gray frame
    cv2.imshow('motion detection steps', numpy_horizontal)

    if cv2.waitKey(1) & 0xFF == ord('q'):
        break

camera.release()
cv2.destroyAllWindows()



# This file contains tests for the motion detector module and aims to verify that the motion detector
# is able to distinguish between videos that contain real motion and videos that do not contain any real motion

# The tests are based on sample videos that are located in the samples directory
# The samples directory contains two subdirectories: positive and negative
# The positive directory contains videos that contain real motion
# The negative directory contains videos that do not contain any real motion, but rather static scenes
# with some noise or light changes

import cv2
import os
from glob import glob
import unittest

from vigi.motion_detector import MotionDetector


class FileProcessor:
    """
    This class is responsible for processing video files and testing them for motion
    if motion is detected, the process method should return True, otherwise it should return False
    """
    def __init__(self, file_path):
        self.file_path = file_path
        self.motion_detected = False
    
    def motion_callback(self):
        self.motion_detected = True

    def process(self):
        cap = cv2.VideoCapture(self.file_path)

        motion_detector = MotionDetector(self.motion_callback)

        while True:
            # read frames from the sample video
            ret, frame = cap.read()
            if not ret: # end of video
                break

            motion_detector.update(frame)


class MotionDetectorTestCase(unittest.TestCase):
    def test_motion_detection(self):
        # negative samples are videos that do not contain any real motion, but rather static scenes
        # with some noise or light changes
        for file in glob(os.path.join(os.path.dirname(__file__), 'samples', 'negative', '*')):
            print(f"File {file} should not contain any motion.")
            processor = FileProcessor(file)
            processor.process()

            self.assertFalse(processor.motion_detected)

        # positive samples are videos that contain real motion
        for file in glob(os.path.join(os.path.dirname(__file__), 'samples', 'positive', '*')):
            print(f"File {file} should contain motion.")
            processor = FileProcessor(file)
            processor.process()

            self.assertTrue(processor.motion_detected)



import numpy as np

def boxes_intersect(box1, box2):
    """
    Check if two bounding boxes (box1 and box2) intersect with each other.
    Returns True if the boxes intersect, False otherwise.
    """

    # Unpack the coordinates
    b1_x1, b1_y1, b1_x2, b1_y2 = box1
    b2_x1, b2_y1, b2_x2, b2_y2 = box2

    # Check for overlap in both x and y axes
    overlap_x = np.minimum(b1_x2, b2_x2) - np.maximum(b1_x1, b2_x1) > 0
    overlap_y = np.minimum(b1_y2, b2_y2) - np.maximum(b1_y1, b2_y1) > 0

    return overlap_x and overlap_y



# This file contains the Database class which is used to interact with the SQLite database

import logging
import sqlite3

class Database:
    def __init__(self, db_path):
        """
        Initialize the database connection
        db_path: str, the path to the database file
        """
        self.db_path = db_path
        self.conn = sqlite3.connect(self.db_path)
        self.conn.row_factory = self.dict_row_factory

    def dict_row_factory(self, cursor, row):
        """
        Convert the row to a dictionary
        cursor: sqlite3.Cursor
        row: sqlite3.Row

        The idea was taken from: https://stackoverflow.com/a/3300514
        """

        result = {}

        for idx, col in enumerate(cursor.description):
            result[col[0]] = row[idx]

        return result

    def init_db(self):
        """
        Initialize the database with the necessary tables and indexes
        create the database file if it does not exist
        create the table if it does not exist
        """
        logging.info(f"Initalizing database at {self.db_path}")
        self.create_table_sql = """
        CREATE TABLE IF NOT EXISTS recordings (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            date TEXT NOT NULL,
            time TEXT NOT NULL,
            camera_id INTEGER NOT NULL,
            tags TEXT
        );
        """
        self.conn.execute(self.create_table_sql)

        self.add_index = """
        CREATE INDEX IF NOT EXISTS idx_recordings ON recordings (date, time, camera_id);
        """
        self.conn.execute(self.add_index)
        logging.info("Database initialized")

    def add_recording(self, date, time, camera_id, tags):
        logging.debug(f"Adding recording to database: {date} {time} {camera_id} {tags}")
        self.conn.execute("INSERT INTO recordings (date, time, camera_id, tags) VALUES (?, ?, ?, ?)", (date, time, camera_id, tags))
        self.conn.commit()

    def find_recording(self, date, time, camera_id):
        """
        Find the recordings for the given date
        date: str, the date in the format YYYY-MM-DD
        time: str, the time in the format HH:MM:SS
        camera_id: int, the id of the camera
        """
        logging.debug(f"Finding recording in database: {date}, {time}, {camera_id}")
        cursor = self.conn.execute("SELECT * FROM recordings WHERE date = ? AND time = ? AND camera_id = ?", [date, time, camera_id])
        return cursor.fetchone()

    def delete_recording(self, date, time, camera_id):
        """
        Delete the recording for the given date, time and camera_id
        date: str, the date in the format YYYY-MM-DD
        time: str, the time in the format HH:MM:SS
        camera_id: int, the id of the camera
        """
        logging.debug(f"Deleting recording from database: {date}, {time}, {camera_id}")
        self.conn.execute("DELETE FROM recordings WHERE date = ? AND time = ? AND camera_id = ?", [date, time, camera_id])
        self.conn.commit()

    def close(self):
        logging.info("Closing database connection")
        self.conn.close()
        self.conn = None

    def integrity_check(self):
        """
        Run the integrity check on the database
        """
        logging.info("Running integrity check on the database")
        self.conn.execute("PRAGMA integrity_check")
        logging.info("Integrity check completed")



# This file contains the EmailNotificationProvider class which is used to send email notifications
import logging

import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

class EmailNotificationProvider:
    def __init__(self, smtp_server: str, smtp_port: int, smtp_user: str, smtp_password: str, sender_email: str,
                 recipient_emails: list):
        """
        :param smtp_server: SMTP server address
        :param smtp_port: SMTP server port
        :param smtp_user: SMTP server user
        :param smtp_password: SMTP server password
        :param sender_email: Email address of the sender
        :param recipient_emails: List of email addresses of the recipients
        """
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.smtp_user = smtp_user
        self.smtp_password = smtp_password
        self.sender_email = sender_email
        self.recipient_emails = recipient_emails
        logging.info(f"EmailNotificationProvider initialized with sender_email: {sender_email}, recipient_emails: {recipient_emails}")

    def notify(self, notification_text: str):
        """
        Send email notification to the recipients
        """
        for recipient_email in self.recipient_emails:
            self._send_email(notification_text, recipient_email)

    def _send_email(self, notification_text: str, recipient_email: str):
        """
        Send email to a single recipient
        """
        logging.info(f"Sending email from {self.sender_email} to {recipient_email} with text: {notification_text}")

        message = MIMEMultipart()
        message["From"] = self.sender_email
        message["To"] = recipient_email
        message["Subject"] = "ViGi Agent Notification"
        message.attach(MIMEText(notification_text, 'plain'))

        with smtplib.SMTP(self.smtp_server, self.smtp_port) as server:
            server.login(self.smtp_user, self.smtp_password)
            server.sendmail(self.sender_email, recipient_email, message.as_string())



import os
import time
import logging
import tempfile
import shutil

import cv2

from vigi.utils.media import generate_preview

DEFAULT_FPS = 25
DEFAULT_FRAME_WIDTH = 640
DEFAULT_FRAME_HEIGHT = 480

class VideoRecorder():
    """
    A class that records the video to a file.
    """
    def __init__(self, recording_path: str, camera_id: int):
        self.recording = False
        self.recording_path = recording_path
        self.camera_id = camera_id
        self.recording_full_path = None

    def start_recording(self, frame_width = None, frame_height = None, fps = None) -> None:
        """
        Start recording the video to a file.
        For each recording session, we pass the set of parameters (frame_width, frame_height, fps),
        as they can be changed between recording sessions.
        """

        logging.info("Starting video recording...")
        logging.info(f"Using parameters: Frame width: {frame_width}, frame height: {frame_height}, FPS: {fps}")

        if frame_width is None or frame_height is None:
            logging.error(f"Frame width and height were not provided. Set to default values of {DEFAULT_FRAME_WIDTH}x{DEFAULT_FRAME_HEIGHT}")
            frame_width = DEFAULT_FRAME_WIDTH
            frame_height = DEFAULT_FRAME_HEIGHT
            return

        if fps is None:
            logging.error(f"FPS was not provided. Set to default value of {DEFAULT_FPS}")
            fps = DEFAULT_FPS
            return

        self.recording = True
        self.recording_start_time = time.strftime("%H-%M-%S")
        self.recording_start_date = time.strftime("%Y-%m-%d")

        # save recording to a temporary directory
        self.recording_full_path = os.path.join(tempfile.gettempdir(), f"camera_{self.camera_id}.mp4")
        logging.info(f"Recording video to: {self.recording_full_path}")

        # fourcc = cv2.VideoWriter_fourcc(*'mp4v') # MPEG-4 Part 2 (Simple Profile) codec, doesn't work in Chrome
        fourcc = cv2.VideoWriter_fourcc(*'avc1') # H.264 codec is not available on Apple M1, but 'AVC1' is the same as 'h264' and works in Chrome
        self.video_writer = cv2.VideoWriter(
            filename = self.recording_full_path,
            fourcc = fourcc,
            fps = int(fps),
            frameSize = (int(frame_width), int(frame_height)),
        )

    def еnd_recording(self) -> None:
        """
        End the recording of the video to a file.
        """
        logging.info("Ending video recording...")
        self.recording = False
        self.video_writer.release()

        # ensure that the recording path exists
        # if not, create the directory
        if not os.path.exists(self.recording_path):
            logging.info(f"Creating directory: {self.recording_path}")
            os.makedirs(self.recording_path)

        camera_id_path = os.path.join(self.recording_path, f"camera_{self.camera_id}")
        if not os.path.exists(camera_id_path):
            logging.info(f"Creating directory: {camera_id_path}")
            os.makedirs(camera_id_path)
        
        # The recordings for each day will be saved in a separate directory
        # Ensure that the date path exists
        # if not, create the directory
        date_path = os.path.join(camera_id_path, self.recording_start_date)
        if not os.path.exists(date_path):
            logging.info(f"Creating directory: {date_path}")
            os.makedirs(date_path)
        
        file_name = f"{self.recording_start_time}.mp4"

        complete_recording_path = os.path.join(date_path, file_name)

        logging.info(f"Moving the recording from {self.recording_full_path} to {complete_recording_path}")
        # move the recording from the temporary directory to the date directory
        shutil.move(self.recording_full_path, complete_recording_path)

        # generate a preview image for the recording
        jpg = generate_preview(complete_recording_path)
        preview_path = os.path.join(date_path, f"{self.recording_start_time}.jpg")
        logging.info(f"Saving preview image to: {preview_path}")
        with open(preview_path, "wb") as f:
            f.write(jpg)


    def is_recording(self) -> bool:
        """
        Returns True if the video is currently being recorded, otherwise False.
        """
        return self.recording

    def add_frame(self, frame) -> None:
        if self.recording:
            # Add the frame to the video file
            self.video_writer.write(frame)
            return



# A script to profile the CPU and memory usage of a program over time
# requires the psutil library to be installed (pip install psutil)
# Usage: sudo python cpu_mem_profiler.py <pid>

import sys
import time

import psutil
import matplotlib.pyplot as plt


# Lists to hold the data
cpu_percentages = []
memory_usage = []

# get pid from command line argument
pid = int(sys.argv[1])

# Get the process
process = psutil.Process(pid)


# Monitoring interval (seconds)
interval = 1
duration = 120  # How long to monitor (seconds)

for _ in range(int(duration / interval)):
    # Record the current CPU and memory utilization
    cpu_percentages.append(process.cpu_percent())
    uss = process.memory_full_info().uss / 1024 /1024 # Convert to Mb

    memory_usage.append(uss)
    
    # Wait for the next interval
    time.sleep(interval)

# Plot two subplots for CPU and memory usage
fig, ax1 = plt.subplots()

color = 'tab:red'
ax1.set_xlabel('Time (seconds)')
ax1.set_ylabel('CPU Usage (%)', color=color)
ax1.plot(cpu_percentages, color=color)
ax1.tick_params(axis='y', labelcolor=color)

ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
color = 'tab:blue'
ax2.set_xlabel('Time (seconds)')
ax2.set_ylabel('Memory Usage (Mb)', color=color)  # we already handled the x-label with ax1
ax2.plot(memory_usage, color=color)
ax2.tick_params(axis='y', labelcolor=color)

fig.tight_layout()  # otherwise the right y-label is slightly clipped
plt.title('CPU and Memory Usage Over Time')
plt.savefig('cpu_mem_profile.png')



class CameraConfig:
    def __init__(self):
        self.camera_id = 0
        self.max_errors = 50
        self.sensitivity = 0.5

    def set_camera_id(self, camera_id):
        # validate the camera id
        camera_id = int(camera_id)
        if camera_id < 0:
            raise ValueError('Camera id must be a positive integer')
        self.camera_id = camera_id

    def set_max_errors(self, max_errors):
        # set the maximum number of consecutive errors when reading a frame from the camera
        max_errors = int(max_errors)
        if max_errors < 0:
            raise ValueError('Max errors must be a positive integer')
        self.max_errors = max_errors

    def set_sensitivity(self, sensitivity):
        # set the sensitivity of the motion detector
        sensitivity = float(sensitivity)
        if sensitivity < 0 or sensitivity > 1:
            raise ValueError('Sensitivity must be between 0 and 1')
        self.sensitivity = sensitivity



# This file is used to create a cache object for the application
# and then import it in the main application file (vigi/__init__.py)
# as well as in the blueprint files (vigi/routes/*.py)

from flask_caching import Cache

cache = Cache(config={'CACHE_TYPE': 'SimpleCache'})



# This file is the entry point of the application. It reads the configuration file
# and the command line arguments, initializes the logger, the notifier, the video recorder
# and the camera monitor, and starts the Flask web server.

import os
import argparse
import logging
import configparser
import atexit

from vigi.configuration_manager import ConfigurationManager

from vigi.notification_providers.email_notification_provider import EmailNotificationProvider
from vigi.notification_providers.sms_notification_provider import SMSNotificationProvider
from vigi.notifier import Notifier

from vigi.app import app
from vigi.video_recorder import VideoRecorder
from vigi.camera_monitor import CameraMonitor

from vigi.database import Database


def read_args():
    """
    read the command line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", help="Enable debug mode", action='store_true')
    parser.add_argument("--no-monitor", help="Disable the camera monitor", action='store_true')
    parser.add_argument("--data-dir", help="Directory to store the recordings", type=str)
    parser.add_argument("--camera-id", help="Camera ID to monitor", type=int)
    parser.add_argument("--host", help="Host to run the web server", type=str)
    parser.add_argument("--port", help="Port to run the web server", type=int)
    parser.add_argument("--max-errors", help="Maximum number of consecutive errors when reading a frame from the camera", type=int)
    parser.add_argument("--sensitivity", help="Sensitivity of the motion detector, should be a float between 0 and 1", type=float)
    parser.add_argument("--detection-model-file", help="Path to the detection model file (YOLO's yolov8n.pt, by default)", type=str)
    args = parser.parse_args()
    return args


def read_config():
    """
    Read the configuration file.
    """
    logging.info("Reading the configuration file... ")
    ini = configparser.ConfigParser()
    ini.read('vigi.ini')
    default = ini['DEFAULT']
    logging.info("Configuration file read successfully.")
    logging.debug(f"Default Configuration: {dict(default)}")

    camera_configs = []
    for section in ini.sections():
        # read the configuration from each camera
        if section.startswith('CAMERA'):
            camera_config = ini[section]
            logging.debug(f"Camera configuration: {dict(camera_config)}")
            camera_configs.append(camera_config)

    return default, camera_configs


def init_logger(debug):
    """
    Initialize the logger.
    """
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)


def init_notifier(configuration_manager):
    """
    Initialize the notifier and its notification providers.
    """
    logging.info("Initializing the notifier... ")
    notification_providers = []
    if configuration_manager.smtp_server_config:
        smtp_server_config = configuration_manager.smtp_server_config
        email_notification_provider = EmailNotificationProvider(
                smtp_server = smtp_server_config['smtpServer'],
                smtp_port = int(smtp_server_config['smtpPort']),
                smtp_user = smtp_server_config['smtpUser'],
                smtp_password = smtp_server_config['smtpPassword'],
                sender_email = smtp_server_config['senderEmail'],
                recipient_emails = smtp_server_config['recipientEmails']
            )
        notification_providers.append(email_notification_provider)

    if configuration_manager.twilio_config:
        twilio_config = configuration_manager.twilio_config
        sms_notification_provider = SMSNotificationProvider(
                account_sid = twilio_config['twilioAccountSid'],
                auth_token = twilio_config['twilioAuthToken'],
                from_number = twilio_config['twilioFromNumber'],
                recipient_phone_numbers = twilio_config['recipientPhoneNumbers']
            )
        notification_providers.append(sms_notification_provider)

    notifier = Notifier(notification_providers)
    logging.info("Notifier initialized successfully.")
    return notifier


def main():
    # Initialize the configuration with the default values
    app.configuration_manager = ConfigurationManager()

    # First, read the configuration file and update the configuration
    default_config, camera_configs = read_config()
    app.configuration_manager.update_from_config(default_config, camera_configs)

    # Then, read the command line arguments and update the configuration
    # the command line arguments take precedence over the configuration file
    args = read_args()
    app.configuration_manager.update_from_args(args)

    # create data dir if it does not exist
    if not os.path.exists(app.configuration_manager.data_dir):
        logging.info(f"Data directory does not exist, creating: {app.configuration_manager.data_dir}")
        os.makedirs(app.configuration_manager.data_dir)

    init_logger(app.configuration_manager.debug)

    notifier = init_notifier(app.configuration_manager)

    logging.info("Initializing the database... ")
    database = Database(app.configuration_manager.db_path)
    database.init_db()
    database.integrity_check()

    # close the database connection after initializing the database 
    # as it's not used in the main thread
    database.close()

    logging.info("Database initialized successfully.")

    if app.configuration_manager.no_monitor:
        logging.info("Camera monitor is disabled.")
    else:
        app.camera_monitors = {}

        for camera_id, camera_config in app.configuration_manager.cameras_config.items():
            logging.info("Initializing the video recorder... ")
            video_recorder = VideoRecorder(
                recording_path = app.configuration_manager.data_dir,
                camera_id=camera_id
            )
            logging.info("Video recorder initialized successfully.")

            # create a camera monitor for each camera
            logging.info(f"Starting the camera monitor for camera {camera_id}... ")
            camera_monitor = CameraMonitor(
                video_recorder = video_recorder,
                camera_id = camera_id,
                max_errors = camera_config.max_errors,
                notifier = notifier,
                db_path = app.configuration_manager.db_path,
                sensitivity=camera_config.sensitivity,
                debug=app.configuration_manager.debug
            )
            camera_monitor.start()
            app.camera_monitors[camera_id] = camera_monitor
            logging.info(f"Camera monitor for camera {camera_id} started successfully.")

    def graceful_exit():
        logging.info("Exiting the application... ")
        if hasattr(app, 'camera_monitors'):
            for camera_monitor in app.camera_monitors.values():
                camera_monitor.stop()
        logging.info("Application exited successfully.")

    atexit.register(graceful_exit)

    logging.info("Starting the Flask web server... ")
    flask_debug = False
    if app.configuration_manager.debug and app.configuration_manager.no_monitor:
        # enable debug mode if the monitor is disabled, because
        # it cause race conditions in multithreading
        flask_debug = True
        logging.warning("Flask debug mode is enabled.")

    app.run(host=app.configuration_manager.host, port=app.configuration_manager.port, debug=flask_debug)



import unittest
from unittest.mock import patch, MagicMock
from vigi.notification_providers.sms_notification_provider import SMSNotificationProvider

class TestSMSNotificationProvider(unittest.TestCase):
    @patch('vigi_agent.notification_providers.sms_notification_provider.Client')
    def test_notify(self, mock_client):
        # Initialize the mocked Twilio client
        mock_client_instance = MagicMock()
        mock_client.return_value = mock_client_instance

        provider = SMSNotificationProvider(
            account_sid = "account_sid",
            auth_token = "auth_token",
            from_number = "+11223344",
            recipient_phone_numbers = ["+1234567890", "+9876543210"]
        )


        provider.notify("Test notification")
        mock_client.assert_called_with("account_sid", "auth_token")

        self.assertEqual(mock_client_instance.messages.create.call_count, 2)



import unittest
from vigi.utils.spatial import boxes_intersect

class SpatialTestCase(unittest.TestCase):
    def test_boxes_intersect(self):
        # boxes_intersect should return True if two boxes intersect
        self.assertTrue(boxes_intersect((0, 0, 2, 2), (1, 1, 3, 3)))

        # boxes_intersect should return False if two boxes do not intersect
        self.assertFalse(boxes_intersect((0, 0, 1, 1), (2, 2, 3, 3)))



# this class is used to validate and store the configuration of the agent
import os

from platformdirs import user_data_dir

from .camera_config import CameraConfig

class ConfigurationManager:
    def __init__(self):
        # default configuration
        self.port = 5000
        self.host = 'localhost'
        self.data_dir = user_data_dir('vigi-agent', 'Vigi')
        self.debug = False
        self.smtp_server_config = None
        self.twilio_config = None
        self.no_monitor = False
        self.db_path = os.path.join(self.data_dir, 'vigi.db')
        self.cameras_config = {}
        self.detection_model_file = './yolov8n.pt'

    def set_port(self, port):
        # validate the port
        port = int(port)
        if port < 0 or port > 65535:
            raise ValueError('Port must be between 0 and 65535')
        self.port = port
    
    def set_host(self, host):
        # validate the host
        if host == '':
            raise ValueError('Host must not be empty')
        self.host = host

    def set_data_dir(self, data_dir):
        # validate the data directory
        if data_dir == '':
            raise ValueError('Data directory must not be empty')
        self.data_dir = data_dir
        self.db_path = os.path.join(self.data_dir, 'vigi.db')

    def set_debug(self, debug):
        # set the debug mode
        self.debug = debug

    def set_detection_model_file(self, detection_model_file):
        if not os.path.exists(detection_model_file):
            raise ValueError('Detection model file does not exist')

        self.detection_model_file = detection_model_file

    def update_from_args(self, cmd_args):        
        # update the configuration from the command line arguments
        if cmd_args.port:
            self.set_port(cmd_args.port)
        if cmd_args.host:
            self.set_host(cmd_args.host)
        if cmd_args.data_dir:
            self.set_data_dir(cmd_args.data_dir)
        if cmd_args.debug:
            self.set_debug(True)
        if cmd_args.detection_model_file:
            self.set_detection_model_file(cmd_args.detection_model_file)
        if cmd_args.no_monitor:
            self.no_monitor = True

        # configure at leas one camera
        camera_id = 0
        if cmd_args.camera_id is not None:
            camera_id = cmd_args.camera_id

        # configure the camera
        camera_config = CameraConfig()
        camera_config.set_camera_id(camera_id)

        if cmd_args.max_errors:
            camera_config.set_max_errors(cmd_args.max_errors)
        if cmd_args.sensitivity:
            cmd_args.set_sensitivity(cmd_args.sensitivity)

        self.cameras_config[camera_id] = camera_config

    def update_from_config(self, default_config, camera_configs):
        # update the configuration from the configuration file
        if 'Port' in default_config:
            self.set_port(default_config['Port'])
        if 'Host' in default_config:
            self.set_host(default_config['Host'])
        if 'DataDir' in default_config:
            self.set_data_dir(default_config['DataDir'])
        if 'Debug' in default_config:
            self.set_debug(default_config['Debug'] == 'True')
        if 'NoMonitor' in default_config:
            self.no_monitor = default_config['NoMonitor'] == 'True'
        if 'DetectionModelFile' in default_config:
            self.set_detection_model_file(default_config['DetectionModelFile'])

        # SMTP configuration
        if 'smtpServer' in default_config:
            self.smtp_server_config = {}

            smtp_params = ['smtpServer', 'smtpPort', 'smtpUser', 'smtpPassword', 'senderEmail']
            for param in smtp_params:
                self.smtp_server_config[param] = default_config[param]    
            
            self.smtp_server_config['recipientEmails'] = default_config['recipientEmails'].split(',')

        if 'twilioAccountSid' in default_config:
            self.twilio_config = {}

            twilio_params = ['twilioAccountSid', 'twilioAuthToken', 'twilioFromNumber']
            for param in twilio_params:
                self.twilio_config[param] = default_config[param]
            
            self.twilio_config['toNumbers'] = default_config['toNumbers'].split(',')

        for camera_config_info in camera_configs:
            camera_config = CameraConfig()
            camera_config.set_camera_id(camera_config_info['CameraId'])

            if 'MaxErrors' in camera_config_info:
                camera_config.set_max_errors(camera_config_info['MaxErrors'])
            
            if 'Sensitivity' in camera_config_info:
                camera_config.set_sensitivity(camera_config_info['Sensitivity'])

            self.cameras_config[camera_config.camera_id] = camera_config



import threading
from datetime import datetime
import time
import logging

import cv2

from .motion_detector import MotionDetector

from vigi.utils.fps_calculator import FPSCalculator
from .utils.pub_sub import PubSub
from .database import Database


class CameraMonitor(threading.Thread):
    """
    A class that monitors the camera for motion and publishes the video stream from the camera.
    """

    def __init__(self, video_recorder = None, camera_id=0, max_errors=50, add_seconds_after_motion=10, notifier=None, db_path=None,
                 sensitivity=0.5, debug=False):
        """
        max_errors: int - the maximum number of consecutive errors when reading a frame from the camera
                            before the camera monitor stops. It's used to prevent the camera monitor from
                            creating a big number of small recordings on each small motion.
        add_seconds_after_motion: int - the number of seconds to add to the video after the motion is detected
        notifier: Notifier - a notifier object to send notifications about the motion
        db_path: str - the path to the database file
        """
        super().__init__()

        self.notifier = notifier

        self.camera_id = camera_id
        self.max_errors = max_errors
        self.add_seconds_after_motion = add_seconds_after_motion

        # Initialize the PubSub object to stream the frames from the camera
        # so other parts of the application can access the stream of frames
        self.frame_stream = PubSub()

        # Initialize the motion detector, when motion is detected, the motion_callback will be called
        self.motion_detector = MotionDetector(
                motion_callback=self.motion_callback,
                sensitivity=sensitivity
            )

        # video_recorder is used to save the video to a file when motion is detected
        self.video_recorder = video_recorder

        # save the start time of the camera monitor to calculate the uptime
        self.start_time = datetime.now()

        self.fps_calculator = FPSCalculator(max_history_size=50)

        # number of frames to add after the motion is over,
        # initially it's set to 0, but it could be assigned a value when the motion is detected
        # and decremented in each frame until it reaches 0
        self.add_frames = 0

        # A flag to stop the camera monitor
        self.should_stop = False

        # A set of detected objects during the recording
        self.detected_objects = set()

        # we can't initialize the database here, because the database should be initialized in
        # the same thread where it's used, but __init__ is called in the main thread
        self.db_path = db_path

    def current_fps(self):
        """
        Returns the current FPS of the system using the FPS calculator if it has calculated the FPS,
        otherwise returns the FPS of the camera.
        """
        fps = self.fps_calculator.current_fps()
        logging.info(f"Calculated FPS = {fps}")
        if fps is None:
            fps = int(self.camera_fps)
            logging.warning(f"FPS is not calculated yet, using the camera's FPS = {fps}")

        # TODO: Set default FPS to 30 if the FPS from the camera is 1 or otherwize invalid

        return fps

    def motion_callback(self):
        logging.info("Motion detected!")

        # send a notification about the motion
        self.notifier.notify(f"Motion was detected by the camera #{self.camera_id}")

        self.add_frames = self.add_seconds_after_motion * self.current_fps()

        if self.video_recorder.is_recording():
            logging.info("Motion detected while the video is being recorded.")
            return
    
        # reset the set of detected objects
        self.detected_objects = set()

        # We need to determine the FPS of the camera to pass it to the video recorder
        # as FPS will be saved as metadata in the video file
        # if FPS will be overestimated, the video will be played faster than it should be
        # if FPS will be underestimated, the video will be played slower than it should be
        # Start recording the video:
        self.video_recorder.start_recording(frame_width=self.frame_width, frame_height=self.frame_height,
                                            fps=self.current_fps())

    def run(self):
        # Connect to the database. We need to connect to the database in the same thread where it's used
        self.database = Database(self.db_path)

        # Initialize the camera with OpenCV
        logging.info("Starting camera monitor... ")
        camera = cv2.VideoCapture(self.camera_id)  # Use 0 for the first webcam
        if camera.isOpened():
            logging.info("Camera opened successfully.")
        else:
            logging.error(f"Camera with ID={self.camera_id} could not be opened.")
            return

        try:
            self.frame_width = camera.get(cv2.CAP_PROP_FRAME_WIDTH)
            self.frame_height = camera.get(cv2.CAP_PROP_FRAME_HEIGHT)
            self.camera_fps = camera.get(cv2.CAP_PROP_FPS)
            logging.info(f"Camera parameters: frame width: {self.frame_width}, frame height: {self.frame_height}, FPS: {self.camera_fps}")

            error_count = 0
            while not self.should_stop:
                success, frame = camera.read()  # Read a frame from the camera
                if not success:
                    # If the camera fails to read a frame:
                    # - Increment the error count
                    # - Print an error message
                    # - Sleep for 1 second
                    #
                    # If the error count reaches the maximum number of consecutive errors:
                    # - print an error message and break the loop

                    error_count += 1
                    if error_count >= self.max_errors:
                        logging.fatal(f"Maximum number of consecutive errors ({self.max_errors}) reached. Exiting.")
                        break

                    logging.error(f"Failed to read a frame from the camera with ID={self.camera_id}")
                    time.sleep(1)
                    continue

                # Reset the error count if a frame is successfully read
                error_count = 0

                # Apply the motion detector to the frame
                frame, detected_objects = self.motion_detector.update(frame)

                # Add the detected objects to the set of detected objects
                self.detected_objects.update(detected_objects)

                # Publish the frame to the frame stream
                self.frame_stream.publish(frame)

                # if motion is not detected anymore and the video is being recorded, then stop the recording
                if not self.motion_detector.is_motion_detected() and self.video_recorder.is_recording():
                    self.add_frames -= 1
                    if self.add_frames <= 0:
                        self.end_recording()

                # here we send all frames to the video recorder, it's up to the video recorder to decide if
                # it should record the frame or not. It could decide to record additional frames before and after
                # the motion is detected
                self.video_recorder.add_frame(frame)
                self.fps_calculator.update()

        finally:
            if self.video_recorder.is_recording():
                self.end_recording()

            # OpenCV cleanup
            logging.info("Releasing the camera...")
            camera.release()

            # close the database connection to save the data
            self.database.close()

    def end_recording(self):
        """
        End the recording of the video to a file.
        """
        self.database.add_recording(date=self.video_recorder.recording_start_date,
                                    time=self.video_recorder.recording_start_time,
                                    camera_id=self.camera_id, tags=','.join(self.detected_objects))

        # send a notification about the detected objects
        self.notifier.notify(f"Moving objects detected: {', '.join(self.detected_objects)} by the camera #{self.camera_id}")

        self.video_recorder.еnd_recording()

    def stop(self):
        """
        Gracefully shutdown the camera monitor.
        """
        logging.info("Shutting down the camera monitor...")
        self.should_stop = True

        # wait for the camera monitor to stop
        self.join()



from queue import Queue
import threading

class PubSub:
    """
    A simple publish-subscribe (PubSub) class that allows multiple subscribers to
    receive messages from a single publisher on a single topic.
    """
    def __init__(self):
        self.subscribers = []
        # Lock to prevent race conditions when accessing the list of
        # subscribers from different threads
        self.lock = threading.Lock()

    def subscribe(self):
        """
        Subscribe to the PubSub object and return a queue to receive messages.
        """
        queue = Queue()
        with self.lock:
            self.subscribers.append(queue)
        return queue

    def unsubscribe(self, subscriber_queue):
        """
        Unsubscribe from the PubSub object.
        """
        with self.lock:
            self.subscribers.remove(subscriber_queue)

    def publish(self, message):
        """
        Publish a message to all the subscribers.
        """
        with self.lock:
            for subscriber in self.subscribers:
                subscriber.put(message, block = False)



import cv2
import numpy as np
from ultralytics import YOLO

from vigi.utils.spatial import boxes_intersect
from vigi.utils.drawing import draw_bboxes, draw_bbox, draw_title

class MotionDetector():
    def __init__(self, motion_callback, debug=False, sensitivity=0.5):
        """
        Initialize the motion detector with the given sensitivity and motion callback.
        The motion callback is called when motion is detected.
        sensitivity: float, the sensitivity of the motion detector, should be between 0 and 1
        """
        self.sensitivity = sensitivity
        self.back_sub = cv2.createBackgroundSubtractorMOG2(
            varThreshold=(50 / self.sensitivity), # the higher the sensitivity, the lower the threshold
            detectShadows=True
        )
        self.motion_callback = motion_callback
        self.motion_detected = False
        self.debug = debug
        self.skip_frames_count = 50 # warming up frames count
        self.yolo = YOLO('yolov8n.pt')

    def is_motion_detected(self) -> bool:
        """
        Returns True if motion is being detected, otherwise False.
        """
        return self.motion_detected

    def update(self, frame):
        """
        Update the motion detector with the current frame and return the frame with the motion detection overlay.
        """
        original_frame = frame.copy()

        # convert the current frame to grayscale
        gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        # update the background model to get the foreground mask
        fg_mask = self.back_sub.apply(gray_frame)

        # Warming up the model by skipping the first few frames
        if self.skip_frames_count > 0:
            self.skip_frames_count -= 1

            draw_title(original_frame, 'WARMING UP')
            return (original_frame, set())

        # threshold the mask, the min_thresh value is set to 100 by default
        # this value roughly impacts the sensitivity of the motion detection
        min_thresh = 50 / self.sensitivity # the higher the sensitivity, the lower the threshold
        _, motion_mask = cv2.threshold(fg_mask, thresh = min_thresh, maxval = 255, type = cv2.THRESH_BINARY)

        # median blur to remove granular noise
        motion_mask = cv2.medianBlur(motion_mask, ksize = 3)

        # morphological operations to fill in holes
        kernel = np.array((15,15), dtype=np.uint8)

        # morphologyEx with MORPH_OPEN is the same as erode followed by dilate
        motion_mask = cv2.morphologyEx(motion_mask, op = cv2.MORPH_OPEN, kernel = kernel, iterations = 1)

        # morphologyEx with MORPH_CLOSE is the same as dilate followed by erode
        motion_mask = cv2.morphologyEx(motion_mask, op = cv2.MORPH_CLOSE, kernel = kernel, iterations = 1)

        # get contours of the moving objects in the frame
        contours, _ = cv2.findContours(motion_mask, mode = cv2.RETR_EXTERNAL, method = cv2.CHAIN_APPROX_SIMPLE)
        detections = []

        cnt_area_thresh = (2500 / self.sensitivity) # the higher the sensitivity, the lower the are threshold
        for cnt in contours:
            area = cv2.contourArea(cnt)
            if area > cnt_area_thresh:
                x, y, w, h = cv2.boundingRect(cnt)
                detections.append([x, y, x + w, y + h])

        detections = np.array(detections)
        
        if len(detections) > 0:
            # Motion detected!
            if self.motion_detected == False:
                self.motion_detected = True
                self.motion_callback()

            self.reset_motion_flag_after = 2 # frames

            if self.debug:
                draw_bboxes(original_frame, detections)
        else:
            if self.motion_detected:
                if self.reset_motion_flag_after > 0:
                    self.reset_motion_flag_after -= 1
                else:
                    self.motion_detected = False

        if self.motion_detected:
            # draw red rectangle on frame
            cv2.rectangle(original_frame, (0, 0), (original_frame.shape[1], original_frame.shape[0]), (0, 0, 255), 10)
            # draw text on frame
            cv2.putText(original_frame, 'MOTION DETECTED', (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)

        detected_objects = set()
        if self.motion_detected:
            results = self.yolo(original_frame, verbose=False)

            for result in results:
                for box, cls in zip(result.boxes.xyxy, result.boxes.cls):
                    label = result.names[int(cls)]
                    detected_objects.update([label])
                    display = False

                    # check if this box intersects with any of the detections
                    for detection in detections:
                        if boxes_intersect(box, detection):
                            display = True

                    if display:
                        draw_bbox(original_frame, box, label=label)

        return (original_frame, detected_objects)



import unittest
from unittest.mock import patch

from vigi.notification_providers.email_notification_provider import EmailNotificationProvider

# # to have a wild test, please provide your own email credentials
# # and uncomment the following lines:
# class TestEmailNotificationProviderWild(unittest.TestCase):
#     def test_notify(self):
#         # notify should send an email to each recipient
#         provider = EmailNotificationProvider(
#             smtp_server = "smtp.gmail.com",
#             smtp_port = 587,
#             smtp_user = "...",
#             smtp_password = "...",
#             sender_email = "...",
#             recipient_emails = ["..."]
#         )
#         provider.notify("Test notification")

class TestEmailNotificationProvider(unittest.TestCase):
     @patch('vigi_agent.notification_providers.email_notification_provider.smtplib.SMTP')
     def test_notify(self, mock_smtp):
        provider = EmailNotificationProvider(
            smtp_server = "smtp.gmail.com",
            smtp_port = 587,
            smtp_user = "user",
            smtp_password = "password",
            sender_email = "sender@example.com",
            recipient_emails = ["recipient1@example.com",
                                "recipient2@example.com"]
        )
        provider.notify("Test notification")
        self.assertEqual(mock_smtp.call_count, 2)  # Verify SMTP was called twice (once for each recipient)



import unittest

from vigi.context_processors import utility_processor

class TestUtilityProcessor(unittest.TestCase):
    def setUp(self):
        self.utilities = utility_processor()

    def test_format_time(self):
        # Test that the format_time function returns the expected result as a time string
        self.assertEqual(self.utilities['format_time']("12-34-56"), "12:34:56")

    def test_format_duration_none(self):
        # Test the format_duration function should return "N/A" if the duration is None
        self.assertEqual(self.utilities['format_duration'](None), "N/A")

    def test_format_duration_seconds(self):
        # Test formatting of duration in seconds
        self.assertEqual(self.utilities['format_duration'](1), "1 second")
        self.assertEqual(self.utilities['format_duration'](2), "2 seconds")

        # Test formatting of duration in minutes (60 seconds)
        self.assertEqual(self.utilities['format_duration'](60), "1 minute")



import cv2
import numpy as np

from flask import Blueprint, current_app, Response

camera_blueprint = Blueprint('camera', __name__)

def generate_frames(camera_monitor):
    """Generate frames from the camera and send them to the client."""

    # For development, we can run flask without the camera monitor
    # because we don't have a camera connected to the development machine
    # and because we need to restart the server every time we change the code
    stream = None
    if camera_monitor:
        stream = camera_monitor.frame_stream.subscribe()

    try:
        while True:
            # Get the frame from the stream
            if stream:
                frame = stream.get()
            else:
                # generate a black frame for the development environment
                frame = np.zeros((360, 640, 3), dtype=np.uint8)

                # Add no signal text
                font = cv2.FONT_HERSHEY_SIMPLEX
                cv2.putText(frame, 'No signal', (50, 100), font, 2, (255, 255, 255), 2, cv2.LINE_AA)

            # Convert the frame to JPEG
            _ret, jpeg = cv2.imencode('.jpg', frame)

            # Convert the JPG to bytes
            frame_bytes = jpeg.tobytes()

            # Yield the frame to the client
            yield (b'--frame\r\n'
                b'Content-Type: image/jpeg\r\n\r\n' + frame_bytes + b'\r\n\r\n')

    except GeneratorExit:
        # If the client disconnects, unsubscribe from the stream
        if stream:
            camera_monitor.frame_stream.unsubscribe(stream)


# route for video streaming
@camera_blueprint.route('/camera/<camera_id>/live')
def camera(camera_id):
    camera_id = int(camera_id)
    camera_monitor = None
    if hasattr(current_app, 'camera_monitors'):
        camera_monitor = current_app.camera_monitors.get(camera_id)

        if camera_monitor is None:
            return Response("Camera monitor not found", status=404)

    return Response(generate_frames(camera_monitor), mimetype='multipart/x-mixed-replace; boundary=frame')



# This file contains the utility function to draw different figures using OpenCV

import cv2
import numpy as np

def draw_bboxes(image, bboxes):
    """
    Draw bounding boxes on the image.
    """
    for bbox in bboxes:
        x1, y1, x2, y2 = bbox
        # select random color
        color = np.random.randint(0, 255, size=3).tolist()
        cv2.rectangle(image, (x1, y1), (x2, y2), color, 2)

def draw_bbox(image, bbox, color=(255, 0, 255), label=None):
    """
    Draw a single bounding box on the image along with the label.
    """
    x1, y1, x2, y2 = bbox
    cv2.rectangle(image, (int(x1), int(y1)), (int(x2), int(y2)), color, 2)

    if label:
        # display label
        cv2.putText(image, label, (int(x1-5), int(y1-5)), cv2.FONT_HERSHEY_SIMPLEX, 1, color, 2)

def draw_title(image, title, color=(0, 0, 255)):
    """
    Draw a title on the image.
    """
    cv2.putText(image, title, (20, 40), cv2.FONT_HERSHEY_SIMPLEX, 1, color, 2)



# The purpose of this script is to test the YOLO model and OpenCV with the camera feed
# outside of the main application code. This is useful for debugging and testing the YOLO model.

import cv2
from ultralytics import YOLO

yolo = YOLO('yolov8n.pt')
camera = cv2.VideoCapture(0)

while True:
    success, frame = camera.read()  # Read a frame from the camera
    if not success:
        break

    results = yolo(frame)
    for result in results:
        for box, cls in zip(result.boxes.xyxy, result.boxes.cls):
            label = result.names[int(cls)]
            x1, y1, x2, y2 = box
            cv2.rectangle(frame, (int(x1), int(y1)), (int(x2), int(y2)), (0, 255, 0), 2)
            cv2.putText(frame, label, (int(x1), int(y1)), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)

    cv2.imshow('frame', frame)

    if cv2.waitKey(1) & 0xFF == ord('q'):
        break

camera.release()
cv2.destroyAllWindows()



import unittest
from vigi.utils.media import read_video_file_meta, generate_preview

class MediaTestCase(unittest.TestCase):
    def test_read_video_file_meta(self):
        # read_video_file_meta should return the correct metadata for a video file
        meta = read_video_file_meta("tests/samples/positive/sample2.mov")
        self.assertEqual(meta["frame_width"], 1620)
        self.assertEqual(meta["frame_height"], 1080)
        self.assertEqual(meta["fps"], 30.0)
        self.assertEqual(meta["frame_count"], 331)
        self.assertEqual(meta["duration"], 11.0)

    def test_generate_preview(self):
        # preview should return the first frame of a video file
        frame = generate_preview("tests/samples/positive/sample2.mov")
        self.assertIsNotNone(frame)



# This file contains the Notifier class which is high level class for sending notifications to different
# notification providers. The providers implement the notify method, such as email, sms, etc.
import logging

class Notifier:
    def __init__(self, providers):
        self.providers = providers

    def notify(self, message):
        logging.info(f"Sending notification: {message}")

        for provider in self.providers:
            provider.notify(message)



from datetime import datetime

from flask import Blueprint, current_app, render_template

live_blueprint = Blueprint('live', __name__)

# route for live
@live_blueprint.route('/live')
def index():
    cameras = []
    if hasattr(current_app, 'camera_monitors'):
        for camera_id, camera_monitor in current_app.camera_monitors.items():
            cameras.append({
                'id': camera_id,
                'frame_width': camera_monitor.frame_width,
                'frame_height': camera_monitor.frame_height,
                'fps': camera_monitor.current_fps(),
                'start_time': camera_monitor.start_time
            })
    else:
        # dummy values for the development environment
        cameras.append({
            'id': 0,
            'frame_width': 640,
            'frame_height': 480,
            'fps': 30,
            'start_time': datetime.now()
        })

    return render_template('live/index.html', cameras=cameras)



import os
import logging
from pathlib import Path
from glob import glob

from flask import Blueprint, redirect, url_for, render_template, current_app, send_file, request
from vigi.utils.media import generate_preview, read_video_file_meta

from vigi.cache import cache
from vigi.database import Database

recordings_blueprint = Blueprint('recordings', __name__)

@recordings_blueprint.route('/recordings/<camera_id>/<date>/<time>/preview')
@cache.cached(timeout=24*60*60) # Cache the preview for 24 hours as it's a unlikely to change
def preview(camera_id, date, time):
    video_path = _video_file_path(camera_id, date, time)

    if not os.path.exists(video_path):
        return "Video not found", 404

    # get the path without the file extension
    video_preview_path = os.path.splitext(video_path)[0] + ".jpg"

    # check if a cached preview exists
    if os.path.exists(video_preview_path):
        # return the cached preview
        with open(video_preview_path, "rb") as f:
            return f.read(), 200, {'Content-Type': 'image/jpeg'}
    else:
        # generate a new preview
        jpg = generate_preview(video_path)

        if jpg is None:
            return(redirect(url_for('static', filename='no_preview_available.jpg')))

        return jpg, 200, {'Content-Type': 'image/jpeg'}


@recordings_blueprint.route('/recordings/<camera_id>/<date>/<time>/video')
def video(camera_id, date, time):
    """
    Returns the video file for the given camera_id, date and time
    """
    video_path = _video_file_path(camera_id, date, time)

    if not os.path.exists(video_path):
        return "Video not found", 404
    
    # if download=true is in the query string, the video will be downloaded
    as_attachment = (request.args.get("download") == "true")

    return send_file(video_path, as_attachment=as_attachment)


# delete recording
@recordings_blueprint.route('/recordings/<camera_id>/<date>/<time>/delete', methods=['DELETE'])
def delete(camera_id, date, time):
    logging.info(f"Deleting recording: {camera_id}/{date}/{time}")
    database = Database(current_app.configuration_manager.db_path)
    video_path = _video_file_path(camera_id, date, time)

    # delete video file
    if os.path.exists(video_path):
        os.remove(video_path)

    # delete meta data from database
    database.delete_recording(date, time, camera_id)
    database.close()

    # return 200
    return "", 200


@recordings_blueprint.route('/recordings')
def index():
    database = Database(current_app.configuration_manager.db_path)
    recording_path = current_app.configuration_manager.data_dir

    recordings = glob(os.path.join(recording_path, "**", "**", "*.mp4"))

    # check if recordings exist
    if len(recordings) == 0:
        return render_template('recordings/index.html', recording_dates=[])

    # get all unique recording dates from file names and sort them
    recording_dates = set([os.path.basename(Path(file).parent) for file in recordings])
    recording_dates = list(recording_dates)
    recording_dates.sort(reverse=True)

    # for each file, get all relevant information and store it in a dictionary
    recordings = {}
    for recording_date in recording_dates:
        recordings[recording_date] = []

        for file in glob(os.path.join(recording_path, "**", recording_date, "*.mp4")):
            camera_id = Path(file).parent.parent.name.split("_")[1]
            time = Path(file).stem
            meta = read_video_file_meta(file)
            duration = 0
            if meta:
                duration = meta.get("duration")

            recording_info = {
                "time": time,
                "duration": duration,
                "camera_id": camera_id,
            }

            # find meta data for the recording in the database
            meta = database.find_recording(recording_date, time, camera_id)
            if meta:
                recording_info["tags"] = meta["tags"]

            recordings[recording_date].append(recording_info)

        # sort recordings by time
        recordings[recording_date].sort(key=lambda x: x["time"], reverse=True)

    database.close()

    return render_template('recordings/index.html',
                           recording_dates=recording_dates,
                           recordings=recordings)


def _video_file_path(camera_id, date, time):
    """
    Returns the absolute path to the video file for the given camera_id, date and time
    """
    recording_path = current_app.configuration_manager.data_dir
    camera_id_dir = f"camera_{camera_id}"
    return os.path.abspath(os.path.join(recording_path, camera_id_dir, date, f"{time}.mp4"))



# This module contains utility functions for working with media files.

import cv2
import random

def read_video_file_meta(video_path):
    """
    Read the metadata of a video file and return it as a dictionary.
    """
    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        return None

    frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fps = cap.get(cv2.CAP_PROP_FPS)
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    duration = int(frame_count / fps)

    cap.release()

    return {
        "frame_width": frame_width,
        "frame_height": frame_height,
        "fps": fps,
        "frame_count": frame_count,
        "duration": duration
    }

def generate_preview(video_path):
    """
    Open video file, select a random frame and return it as a JPEG image for preview.
    """
    cap = cv2.VideoCapture(video_path)
    if not cap.isOpened():
        return None
    
    # get the total number of frames in the video
    frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    # select random frame to start preview
    frame_number = int(frame_count * random.random())

    # select the frame for the preview
    cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number)

    # read the frame
    ret, frame = cap.read()
    if not ret:
        return None
    
    cap.release()

    frame = cv2.resize(frame, (int(frame.shape[1] / 4), int(frame.shape[0] / 4)))

    # convert the frame to JPEG and return it
    _, frame = cv2.imencode('.jpg', frame)

    return frame.tobytes()



# This is the entry point for the Flask application.

from flask_bootstrap import Bootstrap5
from flask import Flask, redirect, url_for

from .cache import cache
from .context_processors import utility_processor

from .routes.live import live_blueprint
from .routes.camera import camera_blueprint
from .routes.recordings import recordings_blueprint

# Initialize the Flask app
app = Flask(__name__, static_url_path='')

# setup the cache
cache.init_app(app)

# configure context processors for views
app.context_processor(utility_processor)

# Bootstrap wrapper for Flask
bootstrap = Bootstrap5(app)

# register the blueprints (sub-apps)
app.register_blueprint(live_blueprint)
app.register_blueprint(camera_blueprint)
app.register_blueprint(recordings_blueprint)

# the root route redirects to the live view
@app.route('/')
def index():
    return redirect(url_for('live.index'))



# This file is used to calculate the FPS of the system to ensure that the video
# metadata is accurate. The FPS is calculated by measuring the time it takes to
# process a frame and then averaging the time over a number of frames.

from collections import deque
import time

class FPSCalculator():
    def __init__(self, max_history_size=100, min_history_size=20):
        """
        the max_history_size is the number of frames to keep in the history used to calculate the FPS,
        the higher the number, the more accurate the FPS will be, but it will also be slower to react to changes
        """
        self.min_history_size = min_history_size
        self.time_history = deque(maxlen=max_history_size)
        self.old_time = time.time()

    def update(self):
        """
        Tic the FPS calculator, this should be called once per frame
        """
        new_time = time.time()
        self.time_history.append(new_time - self.old_time)
        self.old_time = new_time

    def current_fps(self) -> int:
        """
        Returns the current FPS of the system
        """
        if len(self.time_history) < self.min_history_size:
            return None

        return int(1 / (sum(self.time_history) / len(self.time_history)))



# This file contains the context processors for views
# More info: https://flask.palletsprojects.com/en/3.0.x/templating/#context-processors

import datetime
import humanize

def utility_processor():
    def format_time(time):
        # format time as HH:MM:SS
        return time.replace('-', ':')

    def format_date(date):
        # read date as YYYY-MM-DD
        date = datetime.datetime.strptime(date, '%Y-%m-%d')

        # format date as natural language
        return humanize.naturaldate(date)

    def format_duration(duration):
        if duration is None:
            return "N/A"

        # format duration as n seconds, n minutes, n hours, n days, etc.
        return humanize.precisedelta(datetime.timedelta(seconds=duration))

    return dict(
            format_time=format_time,
            format_duration=format_duration,
            format_date=format_date
        )



import unittest
import time

from vigi.utils.fps_calculator import FPSCalculator

class TestFpsCalculator(unittest.TestCase):
    def test_fps_calculator(self):
        # Test the FPS calculator
        fps_calculator = FPSCalculator(max_history_size=50)

        for i in range(100):
            fps_calculator.update()
            time.sleep(0.02)

        self.assertIn(fps_calculator.current_fps(), (42, 43, 44))

    def test_fps_calculator_no_history(self):
        # current_fps should return None if the history is not long enough
        fps_calculator = FPSCalculator(min_history_size=100)

        for i in range(50):
            fps_calculator.update()
            time.sleep(0.01)

        self.assertIsNone(fps_calculator.current_fps())



# This file contains the SMSNotificationProvider class, which is responsible for sending SMS notifications.
import logging

from twilio.rest import Client

class SMSNotificationProvider:
    def __init__(self, account_sid: str, auth_token: str, from_number: str,
                 recipient_phone_numbers: list):
        """
        :param account_sid: Twilio account SID
        :param auth_token: Twilio auth token
        :param from_number: Twilio phone number
        :param recipient_phone_numbers: List of phone numbers of the recipients
        """
        self.from_number = from_number
        self.recipient_phone_numbers = recipient_phone_numbers
        self.client = Client(account_sid, auth_token)
        logging.info("SMS notification provider initialized successfully.")

    def notify(self, notification_text):
        """
        Send SMS notification to the recipients
        """
        for recipient_phone_number in self.recipient_phone_numbers:
            self._send_sms(notification_text, recipient_phone_number)

    def _send_sms(self, notification_text, recipient_phone_number):
        """
        Send SMS to a single recipient
        """
        logging.info(f"Sending SMS from {self.from_number} to {recipient_phone_number} with text: {notification_text}")
        self.client.messages.create(
            body = notification_text,
            from_ = self.from_number,
            to = recipient_phone_number
        )
